unit QRXMain;

{$IfDef Win32}
{$H-}
{$EndIf}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, DB, DBTables, QuickRep, Buttons,
  FileCtrl, QRXUtil, ToolIntf, Grids, Outline;

type
  TOrdString = class(TObject)
   { A string with order number. To use in StringLists }
  private
    FOrder : integer;
    FString : string;
  public
    constructor Create(aOrder:integer; const aString:string);
  end;

  TJoin = class(TObject)
  { Two ordered strings containing link between two tables. }
  private
    FDetailField : TOrdString;
    FMasterField : TOrdString;
  public
    constructor Create(aDetailField, aMasterField : TOrdString);
    destructor Destroy; override;
  end;

  TQRXTable = class(TObject)
  { tables added to report }
  private
    FError : boolean;
    IntTable : TTable; { TTable object for this QRXTable (in interactive phase) }
    FAvailFields : TStringList; { Fields avaialble for report }
    FSelFields : TStringList; { Fields selected for report }
    FMasterTableName : TFileName; { fisical name of the master table }
    FIsMDRelation : boolean; { TRUE if relation with "Master" table is Master/Detail }

    { The following fields are used in the link-tables page }
    FSelIndex : integer; { SelectedIndex }
    FDetailFields : TStringList;
    FMasterFields : TStringList;
    FJoinedFields : TStringList;

    { The following fields are used only in building phase }
    GenTable : TTable; { TTable object for this QRXTable (in generate phase) }
    FDataSource : TDataSource; { DataSource object associated with this table }
    FController : TQRController; { QRController object associated with this table }
    FTop : integer; { Top of Table object in QuickReport form }
  public
    constructor Create(const DatabaseName, TableName:TFileName); virtual;
    destructor Destroy; override;
  end;

  TFrmQRExpt = class(TForm)
    Notebook: TNotebook;
    Image1: TImage;
    PBoxSample: TPaintBox;
    BtnPrev: TBitBtn;
    BtnNext: TBitBtn;
    BtnCancel: TBitBtn;
    Label2: TLabel;
    LBDir: TLabel;
    LBDescDir: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    LBoxFile: TTableListBox;
    CBoxDrive: TAliasComboBox;
    LBoxDirectory: TDirectoryListBox;
    CBoxFilter: TFilterComboBox;
    EditTableName: TEdit;
    LabelMainTable: TLabel;
    SrcLabel: TLabel;
    DstLabel: TLabel;
    BtnInclude: TSpeedButton;
    BtnIncAll: TSpeedButton;
    BtnExclude: TSpeedButton;
    BtnExcAll: TSpeedButton;
    SrcList: TListBox;
    DstList: TListBox;
    BtnUp: TSpeedButton;
    BtnDown: TSpeedButton;
    Label8: TLabel;
    Label9: TLabel;
    CBoxIndex: TComboBox;
    Label13: TLabel;
    Label14: TLabel;
    Panel1: TPanel;
    Label12: TLabel;
    LBoxDetail: TListBox;
    BtnAdd: TButton;
    Label7: TLabel;
    LBoxMaster: TListBox;
    Label11: TLabel;
    LBoxJoined: TListBox;
    BtnDelete: TButton;
    BtnClear: TButton;
    Label10: TLabel;
    EditTitle: TEdit;
    Label15: TLabel;
    CBPageNumber: TCheckBox;
    CBDateTime: TCheckBox;
    Label16: TLabel;
    BtnSearch: TButton;
    OpenDialog: TOpenDialog;
    Panel2: TPanel;
    ImageHeader: TImage;
    BtnClearImage: TButton;
    OutQRXTables: TOutline;
    GroupBox1: TGroupBox;
    RBContinue: TRadioButton;
    RBMaster: TRadioButton;
    LblMD: TLabel;
    RBSimple: TRadioButton;
    LblSimple: TLabel;
    RadioOrientation: TRadioGroup;
    BtnDeleteNode: TButton;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PBoxSamplePaint(Sender: TObject);
    procedure NotebookPageChanged(Sender: TObject);
    procedure BtnNextClick(Sender: TObject);
    procedure BtnPrevClick(Sender: TObject);
    procedure EditTableNameKeyPress(Sender: TObject; var Key: Char);
    procedure EditTableNameChange(Sender: TObject);
    procedure BtnIncludeClick(Sender: TObject);
    procedure BtnExcludeClick(Sender: TObject);
    procedure BtnIncAllClick(Sender: TObject);
    procedure BtnExcAllClick(Sender: TObject);
    procedure CBoxIndexChange(Sender: TObject);
    procedure BtnAddClick(Sender: TObject);
    procedure LBoxFileDblClick(Sender: TObject);
    procedure BtnDeleteClick(Sender: TObject);
    procedure BtnClearClick(Sender: TObject);
    procedure BtnSearchClick(Sender: TObject);
    procedure BtnClearImageClick(Sender: TObject);
    procedure OutQRXTablesClick(Sender: TObject);
    procedure BtnDeleteNodeClick(Sender: TObject);
    procedure BtnCancelClick(Sender: TObject);
    procedure BtnUpClick(Sender: TObject);
    procedure BtnDownClick(Sender: TObject);
    procedure LBoxDetailClick(Sender: TObject);
    procedure CBoxDriveChange(Sender: TObject);
    procedure LBoxDirectoryChange(Sender: TObject);
    procedure CBoxFilterChange(Sender: TObject);
  private
    DrawBitMap : TBitMap;
    SourceBuffer: PChar;
    PictureFileName : TFileName;
    Database : TDataBase;
    FIsSQLBased : boolean;
    procedure SetNavButtons;
    procedure SetDualListButtons;
    procedure SetLinkButtons;
    procedure MoveSelected(List: TCustomListBox; Items: TStrings);
    function GetFirstSelection(List: TCustomListBox): Integer;
    function GetLastSelection(List: TCustomListBox): Integer;
    procedure SetItem(List: TListBox; Index: Integer);
    procedure ChangeNextGlyph(IsCreate:boolean);
    function GetQRXTable(Index:integer):TQRXTable;
    procedure AddBlankChild;
    procedure DestroyNode(Node:TOutlineNode);
    procedure UpdateAvailSelFields;
    procedure QRXToPgSelectFields;
    procedure PgMainTableToQRX;
    procedure QRXToPgMainTable;
    procedure PgLinkTablesToQRX;
    procedure QRXToPgLinkTables;
    procedure QRXToPgLabelMainTable;
    function GetMasterQRXTable(QRXTable:TQRXTable):TQRXTable;
    function GetRealDetailQRXTable(QRXTable:TQRXTable):TQRXTable;
    procedure BuildCBoxIndex;
    procedure BuildLBoxDetail;
    procedure BuildLBoxMaster;
    procedure FillCBoxIndex(QRXTable:TQRXTable);
    procedure FillLBoxDetail(QRXTable:TQRXTable);
    procedure FillLBoxMaster(QRXTable:TQRXTable);
    procedure FillLBoxJoined(QRXTable:TQRXTable);

    procedure Generate(const FileName, FormIdent, UnitIdent: string;
      ToolServices:TIToolServices);
    procedure InitModules(Form:TForm; UnitStream:TMemoryStream;
      const FormIdent, UnitIdent:string);
    procedure DoneModules(Form:TForm; UnitStream:TMemoryStream;
      const FormIdent:string);
    procedure FmtWrite(Stream: TStream; Fmt: PChar;
      const Args: array of const);
    function CheckName(aOwner:TComponent;
      ComponentName:TComponentName):TComponentName;
    function GetIndexFieldNames(QRXTable:TQRXTable):string;
    function GetMasterSource(QRXTable:TQRXTable):TDataSource;
    function GetMasterReport(QRXTable:TQRXTable):TQRController;
    function GetMasterFields(QRXTable:TQRXTable):string;
    procedure MainLoop(Form:TForm; UnitStream:TMemoryStream);
  end;

var
  FrmQrExpt : TFrmQRExpt;

procedure QuickRepExpert(ToolServices: TIToolServices);

implementation

uses Printers, IStreams, Proxies;

{$R *.DFM}

const
  pgSelectMainTable = 0;
  pgSelectFields = 1;
  pgLinkTables = 2;
  pgRadioLinks = 3;
  pgCreate = 4;

  roSimple = 0;
  roMasterDetail = 1;

  dlAddSimple = 0;
  dlAddMasterDetail = 1;
  dlContinue = 3;

  SourceBufferSize = 1024;
  CRLF = #13#10;
  BlankEntry = 'b.l.a.n.k';

  rbIgnore = 0;

procedure CreateOrdStringList(Source:TStrings; Dest:TStrings);
{ Cria uma lista de objetos TOrdString e os insere em Dest. Estes
  objetos sao compostos pelos strings de Source }
var
  F : integer;
  S : TOrdString;
begin
  for F := 0 to Source.Count - 1 do begin
    S := TOrdString.Create(F,Source[F]);
    Dest.AddObject(S.FString,S);
  end;
end;

procedure CreateJoinList(Source:TStrings; Dest:TStrings);
{ Cria uma lista de objetos TJoin e os insere em Dest. Estes
  objetos sao compostos pelos TJoin de source }
var
  F : integer;
  J : TJoin;
begin
  for F := 0 to Source.Count - 1 do begin
    J := TJoin.Create(TJoin(Source.Objects[F]).FDetailField,
      TJoin(Source.Objects[F]).FMasterField);
    Dest.AddObject(J.FDetailField.FString + ' -> ' + J.FMasterField.FString,J);
  end;
end;

procedure InsertOrdered(Item:TOrdString; List: TStrings);
var
  F : integer;
begin
  F := 0;
  while (F < List.Count) and (Item.FOrder > TOrdString(List.Objects[F]).FOrder) do
    Inc(F);
  List.InsertObject(F,Item.FString,Item);
end;

procedure InsertJoin(Item:TJoin; List: TStrings);
var
  F : integer;
begin
  F := 0;
  while (F < List.Count)
  and (Item.FDetailField.FOrder > TJoin(List.Objects[F]).FDetailField.FOrder) do begin
    Inc(F);
  end;
  List.InsertObject(F,Item.FDetailField.FString + ' -> ' + Item.FMasterField.FString,Item);
end;

procedure DestroyObjects(List:TStrings);
var
  F : integer;
begin
  for F := 0 to List.Count - 1 do
    if List.Objects[F] <> nil then begin
      List.Objects[F].Destroy;
      List.Objects[F] := nil;
    end;
end;

constructor TOrdString.Create(aOrder:integer; const aString:string);
begin
  inherited Create;
  FOrder := aOrder;
  FString := aString;
end;

constructor TJoin.Create(aDetailField, aMasterField : TOrdString);
begin
  inherited Create;
  FDetailField := TOrdString.Create(aDetailField.FOrder,aDetailField.FString);
  FMasterField := TOrdString.Create(aMasterField.FOrder,aMasterField.FString);
end;

destructor TJoin.Destroy;
begin
  FDetailField.Destroy;
  FMasterField.Destroy;
  inherited Destroy;
end;

constructor TQRXTable.Create;
begin
  inherited Create;
  IntTable := TTable.Create(nil);
  IntTable.DatabaseName := DataBaseName;
  IntTable.TableName := TableName;
  FError := FALSE;
  try
    IntTable.IndexDefs.Update;
    try
      IntTable.FieldDefs.Update;
      FSelFields := TStringList.Create;
      FAvailFields := TStringList.Create;
      FSelIndex := -1; { none selected }
      FDetailFields := TStringList.Create;
      FMasterFields := TStringList.Create;
      FJoinedFields := TStringList.Create;
      IntTable.GetFieldNames(FAvailFields);
    except
      FError := TRUE;
      raise;
    end;
  except
    FError := TRUE;
    raise;
  end;
end;

destructor TQRXTable.Destroy;
begin
  IntTable.Free;
  FSelFields.Free;
  FAvailFields.Free;
  FDetailFields.Free;
  FMasterFields.Free;
  if FJoinedFields <> nil then
    DestroyObjects(FJoinedFields);
  FJoinedFields.Free;
  inherited Destroy;
end;

procedure TFrmQRExpt.FormCreate(Sender: TObject);
begin
  DrawBitMap := TBitMap.Create;
  PBoxSamplePaint(Self);
  AddBlankChild;
  CBoxDrive.BuildAliasList;
end;

procedure TFrmQRExpt.FormDestroy(Sender: TObject);
begin
  DestroyObjects(LBoxDetail.Items);
  DestroyObjects(LBoxMaster.Items);
  DestroyObjects(LBoxJoined.Items);
  DestroyObjects(SrcList.Items);
  DestroyObjects(DstList.Items);
  DrawBitMap.Free;
end;

procedure TFrmQRExpt.PBoxSamplePaint(Sender: TObject);
begin
  case Notebook.PageIndex of
    pgRadioLinks, pgCreate : begin
      DrawBitMap.Handle := LoadBitmap(HInstance, 'QRX_REPOPTIONS');
      PBoxSample.Canvas.Draw(0, 0, DrawBitmap);
    end;
    pgSelectMainTable: begin
      DrawBitMap.Handle := LoadBitmap(HInstance, 'QRX_MAINTABLE');
      PBoxSample.Canvas.Draw(0, 0, DrawBitmap);
    end;
    pgSelectFields : begin
      DrawBitMap.Handle := LoadBitmap(HInstance, 'QRX_SELFIELDS');
      PBoxSample.Canvas.Draw(0, 0, DrawBitmap);
    end;
    pgLinkTables : begin
      DrawBitMap.Handle := LoadBitmap(HInstance, 'QRX_LINKTABLES');
      PBoxSample.Canvas.Draw(0, 0, DrawBitmap);
    end;
  end; { case }
end;

procedure TFrmQRExpt.PgLinkTablesToQRX;
var
  QRXTable : TQRXTable;

  procedure CopyOrdString(Source:TStrings; Dest:TStringList);
  var
    F : integer;
  begin
    Dest.Clear;
    for F := 0 to Source.Count - 1 do
      Dest.Add(Source[F]);
  end;

  procedure CopyJoin(Source:TStrings; Dest:TStringList);
  var
    F : integer;
    J : TJoin;
  begin
    DestroyObjects(Dest); { Only to TJoin Lists }
    Dest.Clear;
    for F := 0 to Source.Count - 1 do begin
      J := TJoin.Create(TJoin(Source.Objects[F]).FDetailField,
        TJoin(Source.Objects[F]).FMasterField);
      Dest.AddObject('',J);
    end;
  end;

begin
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable <> nil then with QRXTable do begin
    FSelIndex := CBoxIndex.ItemIndex;
    CopyOrdString(LBoxDetail.Items,QRXTable.FDetailFields);
    CopyOrdString(LBoxMaster.Items,QRXTable.FMasterFields);
    CopyJoin(LBoxJoined.Items,QRXTable.FJoinedFields);
  end;
end;

procedure TFrmQRExpt.BuildCBoxIndex;
var
  QRXTable : TQRXTable;
begin
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable <> nil then begin
    CBoxIndex.Clear;
    with QRXTable.IntTable do begin
      if (UpperCase(ExtractFileExt(TableName)) = '.DB') and (TableType = ttDefault)
      or (TableType = ttParadox) then
        CBoxIndex.Items.Add('Primary');
      GetIndexNames(CBoxIndex.Items);
    end;
    CBoxIndex.ItemIndex := 0;
    QRXTable.FSelIndex := 0;
  end
end;

procedure TFrmQRExpt.BuildLBoxDetail;
var
  IndexName, IndexField, IndexFields : string;
  I : integer;
  QRXTable : TQRXTable;
begin
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable <> nil then begin
    with CBoxIndex do
      IndexName := Items[ItemIndex];
    DestroyObjects(LBoxDetail.Items);
    LBoxDetail.Clear;
    QRXTable.FDetailFields.Clear;
    if IndexName = 'Primary' then
      I := 0
    else
      I := QRXTable.IntTable.IndexDefs.IndexOf(IndexName);
    IndexFields := QRXTable.IntTable.IndexDefs[I].Fields;
    while IndexFields <> '' do begin
      I := Pos(';',IndexFields);
      if I = 0 then begin
        IndexField := IndexFields;
        IndexFields := ''; { to exit loop }
      end else begin
        IndexField := Copy(IndexFields,1,I-1);
        System.Delete(IndexFields,1,I);
      end;
      QRXTable.FDetailFields.Add(IndexField);
    end;
    CreateOrdStringList(QRXTable.FDetailFields,LBoxDetail.Items);
  end; { if }
end;

procedure TFrmQRExpt.BuildLBoxMaster;
var
  QRXTable, MasterQRXTable : TQRXTable;
  Node : TOutlineNode;
begin
  Node := OutQRXTables.Items[OutQRXTables.SelectedItem];
  QRXTable := TQRXTable(Node.Data);
  MasterQRXTable := TQRXTable(Node.Parent.Data);
  if (QRXTable <> nil) and (MasterQRXTable <> nil) then begin
    DestroyObjects(LBoxMaster.Items);
    LBoxMaster.Clear;
    QRXTable.FMasterFields.Clear;
    MasterQRXTable.IntTable.GetFieldNames(QRXTable.FMasterFields);
    CreateOrdStringList(QRXTable.FMasterFields,LBoxMaster.Items);
  end; { if }
end;

procedure TFrmQRExpt.FillCBoxIndex(QRXTable:TQRXTable);
begin
  if QRXTable <> nil then begin
    CBoxIndex.Clear;
    with QRXTable.IntTable do begin
      if (UpperCase((ExtractFileExt(TableName))) = '.DB') and (TableType = ttDefault)
      or (TableType = ttParadox) then
        CBoxIndex.Items.Add('Primary');
      GetIndexNames(CBoxIndex.Items);
    end;
    CBoxIndex.ItemIndex := QRXTable.FSelIndex;
  end;
end;

procedure TFrmQRExpt.FillLBoxDetail(QRXTable:TQRXTable);
begin
  if QRXTable <> nil then begin
    DestroyObjects(LBoxDetail.Items);
    LBoxDetail.Clear;
    CreateOrdStringList(QRXTable.FDetailFields,LBoxDetail.Items);
  end; { if }
end;

procedure TFrmQRExpt.FillLBoxMaster(QRXTable:TQRXTable);
begin
  if QRXTable <> nil then begin
    DestroyObjects(LBoxMaster.Items);
    LBoxMaster.Clear;
    CreateOrdStringList(QRXTable.FMasterFields,LBoxMaster.Items);
  end; { if }
end;

procedure TFrmQRExpt.FillLBoxJoined(QRXTable:TQRXTable);
begin
  if QRXTable <> nil then begin
    DestroyObjects(LBoxJoined.Items);
    LBoxJoined.Clear;
    CreateJoinList(QRXTable.FJoinedFields,LBoxJoined.Items);
  end; { if }
end;

procedure TFrmQRExpt.QRXToPgLinkTables;
var
  QRXTable : TQRXTable;
begin { QRXToPgLinkTables }
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable.FSelIndex = -1 then begin { First Time }
    BuildCBoxIndex;
    BuildLBoxDetail;
    BuildLBoxMaster;
    DestroyObjects(LBoxJoined.Items);
    LBoxJoined.Clear;
  end else begin
    FillCBoxIndex(QRXTable);
    FillLBoxDetail(QRXTable);
    FillLBoxMaster(QRXTable);
    FillLBoxJoined(QRXTable);
  end;
  SetLinkButtons;
end;

function TFrmQRExpt.GetMasterQRXTable(QRXTable:TQRXTable):TQRXTable;
var
  Index : longint;
  Node : TOutlineNode;
begin
  Result := nil;
  Index := OutQRXTables.GetDataItem(QRXTable);
  if Index <> 0 then begin
    Node := OutQRXTables.Items[Index];
    if Node <> nil then begin
      Node := Node.Parent;
      if Node <> nil then
        Result := TQRXTable(Node.Data);
    end;
  end;
end;

function TFrmQRExpt.GetRealDetailQRXTable(QRXTable:TQRXTable):TQRXTable;
var
  Index : longint;
  Node, Child : TOutlineNode;
  Found : boolean;
begin
  Node := nil;
  Index := OutQRXTables.GetDataItem(QRXTable);
  if Index <> 0 then begin
    Node := OutQRXTables.Items[Index];
    Index := Node.GetFirstChild;
    Found := FALSE;
    while (Index <> -1) and not Found do begin
      Child := OutQRXTables.Items[Index];
      Index := Node.GetNextChild(Index);
      Found := TQRXTable(Child.Data).FIsMDRelation;
    end;
  end;
  if Found then
    Result := TQRXTable(Child.Data)
  else
    Result := nil;
end;

procedure TFrmQRExpt.QRXToPgLabelMainTable;
var
  Node : TOutlineNode;
  MasterTableName : string;
begin
  if OutQRXTables.SelectedItem = 1 then
    LabelMainTable.Caption := 'Select the main table to use for report'
  else begin
    Node := OutQRXTables.Items[OutQRXTables.SelectedItem];
    MasterTableName := Node.Parent.Text;
    LabelMainTable.Caption := 'Select the table to link with ' + MasterTableName +
      ' table';
  end;
end;

procedure TFrmQRExpt.NotebookPageChanged(Sender: TObject);
begin
  SetNavButtons;
  PBoxSamplePaint(Self);
  case Notebook.PageIndex of
    pgSelectMainTable : begin
      QRXToPgLabelMainTable;
    end;
    pgSelectFields : begin
      QRXToPgSelectFields;
      SetDualListButtons;
    end;
    pgLinkTables : begin
      QRXToPgLinkTables;
    end;
    pgRadioLinks : begin
      OutQRXTablesClick(nil);
    end;
  end; { case }
end;

procedure TFrmQRExpt.SetNavButtons;
begin
  case Notebook.PageIndex of
    pgSelectMainTable : begin
      BtnPrev.Enabled := (OutQRXTables.SelectedItem > 1);
      BtnNext.Enabled := (LBoxFile.Items.IndexOf(EditTableName.Text) <> -1);
    end;
    pgSelectFields : begin
      BtnPrev.Enabled := TRUE;
      BtnNext.Enabled := (DstList.Items.Count > 0);
    end;
    pgRadioLinks : begin
      BtnPrev.Enabled := TRUE;
      BtnNext.Enabled := TRUE;
    end;
  end; { case }
end;

function TFrmQRExpt.GetQRXTable(Index:integer):TQRXTable;
begin
  Result := nil;
  if (Index > 0) and (Index <= OutQRXTables.ItemCount) then
    Result := TQRXTable(OutQRXTables.Items[Index].Data);
end;

procedure TFrmQRExpt.DestroyNode(Node:TOutlineNode);
var
  SubNode : TOutlineNode;
  QRXTable : TQRXTable;
begin
  while Node.HasItems do begin
    SubNode := OutQRXTables.Items[Node.Index+1];
    DestroyNode(SubNode);
  end;
  QRXTable := TQRXTable(Node.Data);
  if QRXTable <> nil then
    QRXTable.Destroy;
  OutQRXTables.Delete(Node.Index);
end;

procedure TFrmQRExpt.PgMainTableToQRX;
{ this function move the contents of the form (in pgSelectMainTable Page)
  into current QRXTable. (Create if it does not exists) }
var
  TableName : TFileName;
  QRXTable : TQRXTable;
  Node : TOutlineNode;
  DatabaseName : string;

  function GetDatabaseName:string;
  var
    Directory : string;
  begin
    if FIsSQLBased then
      Result := CBoxDrive.Text
    else begin
      Directory := UpperCase(CBoxDrive.GetDirectory(CBoxDrive.Text));
      if UpperCase(LBoxDirectory.Directory) = Directory then
        Result := CBoxDrive.Text
      else
        Result := LBoxDirectory.Directory;
    end;
  end;

begin
  DatabaseName := GetDatabaseName;
  TableName := DatabaseName + '\' + EditTableName.Text;
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable <> nil then begin { index already used }
    if QRXTable.IntTable.DatabaseName + '\' + QRXTable.IntTable.TableName <> TableName then begin
      DestroyNode(OutQRXTables.Items[OutQRXTables.SelectedItem]);
      QRXTable := nil;
      AddBlankChild;
    end;
  end;
  if QRXTable = nil then begin
    try
      QRXTable := TQRXTable.Create(DatabaseName, EditTableName.Text);
      OutQRXTables.Items[OutQRXTables.SelectedItem].Text := UpperCase(QRXTable.IntTable.TableName);
      OutQRXTables.Items[OutQRXTables.SelectedItem].Data := QRXTable;
    except
      OutQRXTables.Delete(OutQRXTables.SelectedItem);
      AddBlankChild;
      raise;
    end;
  end;
end;

procedure TFrmQRExpt.QRXToPgMainTable;
{ this function move the contents of the current QRXTable into
  form (in pgSelectMainTable Page) }
var
  QRXTable : TQRXTable;
  I : integer;
begin
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable <> nil then begin
    I := CBoxDrive.AliasList.IndexOf(QRXTable.IntTable.DatabaseName);
    if I <> -1 then begin { using alias }
      CBoxDrive.ItemIndex := I;
      CBoxDriveChange(CBoxDrive);
    end else
      LBoxDirectory.Directory := QRXTable.IntTable.DatabaseName;
    I := LBoxFile.Items.IndexOf(QRXTable.IntTable.TableName);
    LBoxFile.ItemIndex := I;
    EditTableName.Text := QRXTable.IntTable.TableName;
  end;
end;

procedure TFrmQRExpt.QRXToPgSelectFields;
var
  QRXTable : TQRXTable;
begin
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  if QRXTable <> nil then begin
    DestroyObjects(SrcList.Items); SrcList.Clear;
    DestroyObjects(DstList.Items); DstList.Clear;
    CreateOrdStringList(QRXTable.FAvailFields,SrcList.Items);
    CreateOrdStringList(QRXTable.FSelFields,DstList.Items);
  end;
end;

procedure TFrmQRExpt.ChangeNextGlyph(IsCreate:boolean);
begin
  if IsCreate then begin
    BtnNext.Glyph.Handle := LoadBitMap(HInstance,'QRX_CREATE');
    BtnNext.Caption := '&Create';
  end else begin
    BtnNext.Glyph.Handle := LoadBitMap(HInstance,'QRX_NEXT');
    BtnNext.Caption := '&Next';
  end;
end;

procedure TFrmQRExpt.AddBlankChild;
{ Add a blank child to current selected item in outline }
var
  QRXTable : TQRXTable;
begin
  with OutQRXTables do begin
    AddChildObject(SelectedItem,BlankEntry,nil);
    FullExpand;
    SelectedItem := GetTextItem(BlankEntry);
  end;
end;

procedure TFrmQRExpt.BtnNextClick(Sender: TObject);
var
  Page : integer;
begin
  Page := Notebook.PageIndex;
  case Page of
    pgSelectMainTable : begin
      PgMainTableToQRX;
      if OutQRXTables.SelectedItem > 1 then
        GetQRXTable(OutQRXTables.SelectedItem).FIsMDRelation := (RBMaster.Checked);
    end;
    pgLinkTables : begin
      PgLinkTablesToQRX;
    end;
    pgRadioLinks : begin
    end;
    pgCreate : begin
      ModalResult := mrOk;
      Exit;
      (*ShowMessage('Atencao. Esta e a versao para depuracao');
      Generate('Teste','Form1','Unit1',nil);*)
    end;
  end; { case }
  if (Page = pgRadioLinks) and not RBContinue.Checked then begin
    AddBlankChild;
    Page := pgSelectMainTable;
  end else
    Page := Page + 1;
  if (Page = pgLinkTables) and (OutQRXTables.SelectedItem = 1) then
    Page := Page + 1; { ignore the LinkFields page }
  if Page = pgSelectMainTable then
    QRXToPgMainTable;
  if Page = pgCreate then
    ChangeNextGlyph(TRUE);
  Notebook.PageIndex := Page;
end;

procedure TFrmQRExpt.BtnPrevClick(Sender: TObject);
var
  Page : integer; { to avoid flicker }
  QRXTable : TQRXTable;
  Index : integer;
begin
  Page := Notebook.PageIndex;
  case Page of
    pgSelectMainTable : begin
      with OutQRXTables do if SelectedItem > 1 then begin
        Index := GetTextItem(BlankEntry);
        if Index > 0 then
          DestroyNode(Items[Index]);
        NoteBook.PageIndex := pgRadioLinks;
        Exit;
      end;
    end;
    pgRadioLinks : begin
    end;
    pgLinkTables : begin
      PgLinkTablesToQRX;
    end;
  end; { case }
  Page := Page - 1;
  case Page of
    pgSelectMainTable : begin
      QRXToPgMainTable;
    end;
  end; { case }
  if (Page = pgLinkTables) and (OutQRXTables.SelectedItem = 1) then
    Page := Page - 1; { ignore the LinkFields page }
  if Page <> pgCreate then
    ChangeNextGlyph(FALSE);
  NoteBook.PageIndex := Page;
end;

procedure TFrmQRExpt.EditTableNameKeyPress(Sender: TObject;
  var Key: Char);
begin
  if Key = #13 then begin
    if FIsSQLBased then
      LBoxFile.BuildTableList(CBoxDrive.Text,EditTableName.Text)
    else
      LBoxFile.ApplyFilePath(EditTableName.Text);
    Key := #0;
  end;
end;

procedure TFrmQRExpt.EditTableNameChange(Sender: TObject);
begin
  SetNavButtons;
end;

procedure TFrmQRExpt.CBoxDriveChange(Sender: TObject);

  procedure RestoreToLocal;
  begin
    CBoxFilter.Filter := 'All Tables (*.DB;*.DBF)|*.DB;*.DBF|Paradox Tables (*.DB)|*.DB|DBase Tables (*.DBF)|*.DBF';
    if FIsSQLBased then
      LBoxDirectory.Update;
    FIsSQLBased := FALSE;
  end;

begin
  if CBoxDrive.ShowingAlias then begin
    Database := Session.OpenDatabase(CBoxDrive.Text);
    try
      if Database.IsSQLBased then begin
        FIsSQLBased := TRUE;
        LBoxDirectory.Items.Clear;
        LBoxDirectory.Items.AddObject(CBoxDrive.Text,CBoxDrive.AliasBitMap);
        EditTableName.Text := '*.*';
        CBoxFilter.Filter := '<User Tables>|*.*|<All Tables>|*.*|';
        LBoxFile.BuildTableList(CBoxDrive.Text,'*.*');
      end else begin
        LBoxDirectory.Directory := CBoxDrive.GetDirectory(CBoxDrive.Text);
        RestoreToLocal;
      end;
    finally
      Session.CloseDatabase(Database);
    end;
  end else begin
    LBoxDirectory.Drive := CBoxDrive.Text[1];
    RestoreToLocal;
  end;
  if FIsSQLBased then begin
    LBDescDir.Visible := FALSE;
    LBDir.Enabled := FALSE;
  end else begin
    LBDescDir.Visible := TRUE;
    LBDir.Enabled := TRUE;
  end;
end;

procedure TFrmQRExpt.LBoxDirectoryChange(Sender: TObject);
begin
  if not FIsSQLBased then
    LBoxFile.Directory := LBoxDirectory.Directory;
end;

procedure TFrmQRExpt.CBoxFilterChange(Sender: TObject);
var
  S : string;

  procedure Update;
  begin
    LBoxFile.BuildTableList(CBoxDrive.Text,EditTableName.Text);
  end;

begin
  S := CBoxFilter.Items[CBoxFilter.ItemIndex];
  if S = '<User Tables>' then begin
    LBoxFile.SystemTables := FALSE;
    Update;
  end else if S = '<All Tables>' then begin
    LBoxFile.SystemTables := TRUE;
    Update;
  end;
end;

(*************************{Routines from DualListBox}********************)
procedure TFrmQRExpt.BtnIncludeClick(Sender: TObject);
var
  Index: Integer;
begin
  Index := GetFirstSelection(SrcList);
  MoveSelected(SrcList, DstList.Items);
  SetItem(SrcList, Index);
  UpdateAvailSelFields;
end;

procedure TFrmQRExpt.BtnExcludeClick(Sender: TObject);
var
  Index: Integer;
begin
  Index := GetFirstSelection(DstList);
  MoveSelected(DstList, SrcList.Items);
  SetItem(DstList, Index);
  UpdateAvailSelFields;
end;

procedure TFrmQRExpt.BtnIncAllClick(Sender: TObject);
var
  I: Integer;
begin
  for I := 0 to SrcList.Items.Count - 1 do
    InsertOrdered(SrcList.Items.Objects[I] as TOrdString,DstList.Items);
  SrcList.Items.Clear;
  SetItem(SrcList, 0);
  UpdateAvailSelFields;
end;

procedure TFrmQRExpt.BtnExcAllClick(Sender: TObject);
var
  I: Integer;
begin
  for I := 0 to DstList.Items.Count - 1 do
    InsertOrdered(DstList.Items.Objects[I] as TOrdString,SrcList.Items);
  DstList.Items.Clear;
  SetItem(DstList, 0);
  UpdateAvailSelFields;
end;

procedure TFrmQRExpt.BtnUpClick(Sender: TObject);
var
  I, J : integer;
  NotSelected : TOrdString;
  TmpItens : TStrings;
begin
  I := GetFirstSelection(DstList);
  if (I <> LB_ERR) and (I > 0) then begin
    TmpItens := TStringList.Create;
    try
      TmpItens.Assign(DstList.Items);
      for J := I to TmpItens.Count-1 do
        if DstList.Selected[J] then begin
          with TOrdString(TmpItens.Objects[J]) do
            FString := '|' + FString; { saving selected }
          if not DstList.Selected[J-1] then
            NotSelected := TOrdString(TmpItens.Objects[J-1]);
          TmpItens.Move(J,J-1);
        end else if (DstList.Selected[J-1]) and (NotSelected <> nil) then begin
          TmpItens.Delete(J-1);
          TmpItens.InsertObject(J-1,NotSelected.FString,NotSelected);
          NotSelected := nil;
        end;
    finally
      DstList.Items.Assign(TmpItens);
      for J := 0 to DstList.Items.Count - 1 do begin
        if TOrdString(DstList.Items.Objects[J]).FString[1] = '|' then begin
          DstList.Selected[J] := TRUE;
          System.Delete(TOrdString(DstList.Items.Objects[J]).FString,1,1);
        end;
      end;
      TmpItens.Free;
    end;
  end; { if }
  UpdateAvailSelFields;
end;

procedure TFrmQRExpt.BtnDownClick(Sender: TObject);
var
  I, J : integer;
  NotSelected : TOrdString;
  TmpItens : TStrings;
begin
  I := GetLastSelection(DstList);
  if (I <> LB_ERR) and (I < DstList.Items.Count - 1) then begin
    TmpItens := TStringList.Create;
    try
      TmpItens.Assign(DstList.Items);
      for J := I downto 0 do
        if DstList.Selected[J] then begin
          with TOrdString(TmpItens.Objects[J]) do
            FString := '|' + FString; { saving selected }
          if not DstList.Selected[J+1] then
            NotSelected := TOrdString(TmpItens.Objects[J+1]);
          TmpItens.Move(J,J+1);
        end else if (DstList.Selected[J+1]) and (NotSelected <> nil) then begin
          TmpItens.Delete(J+1);
          TmpItens.InsertObject(J+1,NotSelected.FString,NotSelected);
          NotSelected := nil;
        end;
    finally
      DstList.Items.Assign(TmpItens);
      for J := 0 to DstList.Items.Count - 1 do begin
        if TOrdString(DstList.Items.Objects[J]).FString[1] = '|' then begin
          DstList.Selected[J] := TRUE;
          System.Delete(TOrdString(DstList.Items.Objects[J]).FString,1,1);
        end;
      end;
      TmpItens.Free;
    end;
  end; { if }
  UpdateAvailSelFields;
end;

procedure TFrmQRExpt.MoveSelected(List: TCustomListBox; Items: TStrings);
var
  I: Integer;
begin
  for I := List.Items.Count - 1 downto 0 do
    if List.Selected[I] then begin
      InsertOrdered(List.Items.Objects[I] as TOrdString,Items);
      List.Items.Delete(I);
    end;
end;

procedure TFrmQRExpt.SetDualListButtons;
var
  SrcEmpty, DstEmpty: Boolean;
begin
  SrcEmpty := SrcList.Items.Count = 0;
  DstEmpty := DstList.Items.Count = 0;
  BtnInclude.Enabled := not SrcEmpty;
  BtnIncAll.Enabled := not SrcEmpty;
  BtnExclude.Enabled := not DstEmpty;
  BtnExcAll.Enabled := not DstEmpty;
  SetNavButtons;
end;

procedure TFrmQRExpt.SetLinkButtons;
begin
  BtnClear.Enabled := (LBoxJoined.Items.Count > 0);
  BtnDelete.Enabled := (LBoxJoined.ItemIndex >= 0);
  BtnNext.Enabled := BtnClear.Enabled;
  BtnAdd.Enabled := (LBoxDetail.ItemIndex >= 0) and (LBoxMaster.ItemIndex >= 0);
end;

function TFrmQRExpt.GetFirstSelection(List: TCustomListBox): Integer;
begin
  for Result := 0 to List.Items.Count - 1 do
    if List.Selected[Result] then Exit;
  Result := LB_ERR;
end;

function TFrmQRExpt.GetLastSelection(List: TCustomListBox): Integer;
begin
  for Result := List.Items.Count - 1 downto 0 do
    if List.Selected[Result] then Exit;
  Result := LB_ERR;
end;

procedure TFrmQRExpt.SetItem(List: TListBox; Index: Integer);
var
  MaxIndex: Integer;
begin
  with List do begin
    SetFocus;
    MaxIndex := List.Items.Count - 1;
    if Index = LB_ERR then Index := 0
    else if Index > MaxIndex then Index := MaxIndex;
    Selected[Index] := True;
  end;
  SetDualListButtons;
end;

procedure TFrmQRExpt.UpdateAvailSelFields;
var
  QRXTable : TQRXTable;
  F : integer;
begin
  QRXTable := GetQRXTable(OutQRXTables.SelectedItem);
  DestroyObjects(QRXTable.FAvailFields); QRXTable.FAvailFields.Clear;
  DestroyObjects(QRXTable.FSelFields); QRXTable.FSelFields.Clear;
  for F := 0 to SrcList.Items.Count - 1 do
    QRXTable.FAvailFields.Add(SrcList.Items[F]);
  for F := 0 to DstList.Items.Count - 1 do
    QRXTable.FSelFields.Add(DstList.Items[F]);
end;


procedure TFrmQRExpt.CBoxIndexChange(Sender: TObject);
begin
  BuildLBoxDetail;
  BuildLBoxMaster;
  DestroyObjects(LBoxJoined.Items);
  LBoxJoined.Clear;
end;

procedure TFrmQRExpt.LBoxFileDblClick(Sender: TObject);
begin
  if BtnNext.Enabled then
    BtnNext.Click;
end;

procedure TFrmQRExpt.BtnCancelClick(Sender: TObject);
begin
  if MessageDlg('Are you sure that you want to close the QuickReport Expert?',
    mtConfirmation,[mbyes,mbno],0) <> mrYes then
      ModalResult := mrNone;
end;

procedure TFrmQRExpt.LBoxDetailClick(Sender: TObject);
begin
  SetLinkButtons;
end;

procedure TFrmQRExpt.BtnAddClick(Sender: TObject);
var
  Join : TJoin;
begin
  Join := TJoin.Create(TOrdString(LBoxDetail.Items.Objects[LBoxDetail.ItemIndex]),
    TOrdString(LBoxMaster.Items.Objects[LBoxMaster.ItemIndex]));
  InsertJoin(Join,LBoxJoined.Items);
  with LBoxDetail do begin
    Items.Objects[ItemIndex].Destroy;
    Items.Delete(ItemIndex);
  end;
  with LBoxMaster do begin
    Items.Objects[ItemIndex].Destroy;
    Items.Delete(ItemIndex);
  end;
  SetLinkButtons;
end;

procedure TFrmQRExpt.BtnDeleteClick(Sender: TObject);
var
  DetailField, MasterField : TOrdString;
begin
  with TJoin(LBoxJoined.Items.Objects[LBoxJoined.ItemIndex]) do begin
    DetailField := TOrdString.Create(FDetailField.FOrder,FDetailField.FString);
    MasterField := TOrdString.Create(FMasterField.FOrder,FMasterField.FString);
  end;
  InsertOrdered(DetailField, LBoxDetail.Items);
  InsertOrdered(MasterField, LBoxMaster.Items);
  with LBoxJoined do begin
    Items.Objects[ItemIndex].Destroy;
    Items.Delete(ItemIndex);
  end;
  SetLinkButtons;
end;

procedure TFrmQRExpt.BtnClearClick(Sender: TObject);
begin
  BuildLBoxDetail;
  BuildLBoxMaster;
  DestroyObjects(LBoxJoined.Items);
  LBoxJoined.Clear;
  SetLinkButtons;
end;

procedure TFrmQRExpt.BtnSearchClick(Sender: TObject);
begin
  if OpenDialog.Execute then begin
    PictureFileName := OpenDialog.FileName;
    ImageHeader.Picture.LoadFromFile(PictureFileName);
  end;
end;

procedure TFrmQRExpt.BtnClearImageClick(Sender: TObject);
begin
  ImageHeader.Picture.Graphic := nil;
  ImageHeader.Invalidate;
  PictureFileName := '';
end;

procedure TFrmQRExpt.OutQRXTablesClick(Sender: TObject);
begin
  LblSimple.Caption := OutQRXTables.Items[OutQRXTables.SelectedItem].Text;
  LblMD.Caption := LBLSimple.Caption;
  BtnDeleteNode.Enabled := (OutQRXTables.SelectedItem > 1);
end;

procedure TFrmQRExpt.BtnDeleteNodeClick(Sender: TObject);
var
  TableName : string;
begin
  with OutQRXTables do
    TableName := Items[SelectedItem].Text;
  if MessageDlg('Are you sure that you want to delete ' +
    TableName + ' table and all its detail tables?',mtConfirmation, [mbYes,mbNo],0) = mrYes then begin
    DestroyNode(OutQRXTables.Items[OutQRXTables.SelectedItem]);
    OutQRXTablesClick(nil);
  end;
end;

(*************************{Routines of generate phase}*********************)
procedure TFrmQRExpt.InitModules(Form:TForm; UnitStream:TMemoryStream;
  const FormIdent, UnitIdent:string);
{ Cria formulario, Unit e Insere Cabecalho nelas }
begin
  with Form do begin
    Left := 217;
    Top := 62;
    Width := 652;
    Height := 437;
    HorzScrollBar.Range := 1200;
    VertScrollBar.Range := 2000;
    AutoScroll := False;
    Caption := 'QuickReport Form';
    Font.Color := clBlack;
    Font.Height := -13;
    Font.Name := 'Arial';
    Font.Style := [];
    PixelsPerInch := 96;
    Scaled := False;
    ShowHint := True;
  end;
  FmtWrite(UnitStream,
    'unit %s;' + CRLF + CRLF +
    'interface' + CRLF + CRLF +
    'uses'#13#10 +
    '  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,'#13#10 +
    '  StdCtrls, ExtCtrls, Forms;' + CRLF + CRLF, [UnitIdent]);
  { begin the class declaration }
  FmtWrite(UnitStream,
    'type'#13#10 +
    '  T%s = class(TForm)'#13#10, [FormIdent]);
end;

procedure TFrmQRExpt.DoneModules(Form:TForm; UnitStream:TMemoryStream;
  const FormIdent:string);
begin
  FmtWrite(UnitStream,'  end;' + CRLF + CRLF + 'var' + CRLF +
    '  %s: T%s;' + CRLF + CRLF + 'implementation' + CRLF + CRLF +
    '{$R *.DFM}' + CRLF + CRLF, [FormIdent, FormIdent]);
  FmtWrite(UnitStream, 'end.' + CRLF, [nil]);
end;

function TFrmQRExpt.CheckName(aOwner:TComponent;
  ComponentName:TComponentName):TComponentName;
var
  Count : integer;
begin
  Count := 2; Result := ComponentName;
  while aOwner.FindComponent(Result) <> nil do begin
    Result := ComponentName + IntToStr(Count);
    Inc(Count);
  end;
end;

function TFrmQRExpt.GetIndexFieldNames(QRXTable:TQRXTable):string;
var
  F, P : integer;
begin
  Result := '';
  for F := 0 to QRXTable.FJoinedFields.Count - 1 do begin
    Result := TJoin(QRXTable.FJoinedFields.Objects[F]).FDetailField.FString + ';';
  end;
  if Result <> '' then
    Dec(Result[0]);
end;

function TFrmQRExpt.GetMasterSource(QRXTable:TQRXTable):TDataSource;
var
  MasterQRXTable : TQRXTable;
begin
  Result := nil;
  MasterQRXTable := GetMasterQRXTable(QRXTable);
  if (MasterQRXTable <> nil) and (MasterQRXTable <> QRXTable)
  and (MasterQRXTable.GenTable <> nil) then
    Result := MasterQRXTable.FDataSource;
end;

function TFrmQRExpt.GetMasterReport(QRXTable:TQRXTable):TQRController;
var
  MasterQRXTable : TQRXTable;
begin
  Result := nil;
  MasterQRXTable := GetMasterQRXTable(QRXTable);
  if (MasterQRXTable <> nil) and (MasterQRXTable.GenTable <> nil) then
    Result := MasterQRXTable.FController;
end;

function TFrmQRExpt.GetMasterFields(QRXTable:TQRXTable):string;
var
  F : integer;
begin
  Result := '';
  for F := 0 to QRXTable.FJoinedFields.Count - 1 do begin
    Result := TJoin(QRXTable.FJoinedFields.Objects[F]).FMasterField.FString + ';';
  end;
  if Result <> '' then
    Dec(Result[0]);
end;

procedure TFrmQRExpt.MainLoop(Form:TForm; UnitStream:TMemoryStream);
const
  IniLeft = 4;
  IniTop = 1;
  Interval = 4;
var
  QuickReport : TQuickReport;
  CurrQRController : TQRController;
  CurrQRXTable, DetailQRXTable : TQRXTable;
  CurrBand : TQRBand;
  F : integer;
  HasMD : boolean;
  aNonVisualTop : integer; { Top para os componentes nao visuais }
  aDesignInfo : longint;
  WorkCanvas : TControlCanvas; { Usado para calculo do tamanho da fonte }
  LabelFont : TFont; { idem }
  FieldsInBand, FieldsInDetailBand : TStringList;
  aBandType : TQRBandType;

  function GetBandWidth:integer;
  { Necessario porque QRBand.Width e' sempre 1200, independente de qualquer
    coisa. Esta funcao retorna o Width disponivel (para eu calcular a
    disposicao dos campos) em funcao da orientacao (Portrait ou Landscape) }
  begin
    if RadioOrientation.ItemIndex = 0 then
      Result := QRPrinter.PageWidth
    else
      Result := QRPrinter.PageHeight;
  end;

  function GetTextWidth(const S:string; aFont:TFont):integer;
  begin
    WorkCanvas.Font.Assign(aFont);
    Result := WorkCanvas.TextWidth(S);
  end;

  function InsertTable(QRXTable:TQRXTable; aTop:integer):TTable;
  var
    DataSource : TDataSource;
  begin
    Result := TTable.Create(Form);
    QRXTable.GenTable := Result;
    with Result do begin
      DataBaseName := QRXTable.IntTable.DatabaseName;
      IndexFieldNames := GetIndexFieldNames(QRXTable);
      TableName := FirstUpCase(QRXTable.IntTable.TableName);
      MasterSource := GetMasterSource(QRXTable);
      MasterFields := GetMasterFields(QRXTable);
      Name := CheckName(Form,'Table' + RemoveExt(TableName));
      LongRec(aDesignInfo).Lo := 36;
      LongRec(aDesignInfo).Hi := aTop;
      Result.DesignInfo := aDesignInfo;
      QRXTable.FTop := aTop;
      Active := TRUE;
    end;
    FmtWrite(UnitStream, '    %s:TTable;' + CRLF, [Result.Name]);

    DataSource := TDataSource.Create(Form);
    QRXTable.FDataSource := DataSource;
    with DataSource do begin
      DataSet := Result;
      LongRec(aDesignInfo).Lo := 68;
      LongRec(aDesignInfo).Hi := aTop;
      DataSource.DesignInfo := aDesignInfo;
      Name := CheckName(Form,'DataSource' + RemoveExt(Result.TableName));
    end;
    FmtWrite(UnitStream, '    %s:TDataSource;' + CRLF, [DataSource.Name]);
  end;

  function InsertQuickRep(QRXTable:TQRXTable):TQuickReport;
  begin
    Result := TQuickReport.Create(Form);
    with Result do begin
      Name := CheckName(Form,'QuickReport');
      DataSource := QRXTable.FDataSource;
      ColumnMarginInches := 0;
      ColumnMarginMM := 0;
      Columns := 1;
      DisplayPrintDialog := False;
      LeftMarginInches := 0;
      LeftMarginMM := 0;
      if RadioOrientation.ItemIndex = 0 then
        Orientation := poPortrait
      else
        Orientation := poLandscape;
{$IfNDef Win32}
      PageFrame.DrawTop := False;
      PageFrame.DrawBottom := False;
      PageFrame.DrawLeft := False;
      PageFrame.DrawRight := False;
      PaperLength := 0;
      PaperSize := qrpDefault;
      PaperWidth := 0;
      PrintOrder := qrColByCol;
{$EndIf}
      ReportTitle := EditTitle.Text;
      RestartData := True;
      SQLCompatible := False;
      TitleBeforeHeader := False;
      LongRec(aDesignInfo).Lo := 4;
      LongRec(aDesignInfo).Hi := QRXTable.FTop;
      Result.DesignInfo := aDesignInfo;
    end;
    FmtWrite(UnitStream, '    QuickReport:TQuickReport;' + CRLF, [nil]);
    QRXTable.FController := Result;
  end;

  function InsertHeader:TQRBand;
  var
    QRSysData : TQRSysData;
    S : string;
    Image : TImage;
  begin
    Result := TQRBand.Create(Form);
    with Result do begin
      Parent := Form;
      Name := CheckName(Form,'Title');
      Left := 0;
      Top := 0;
      Width := 1200;
      Height := 47;
      Align := alTop;
      BandType := rbTitle;
      Color := clWhite;
      Font.Color := clNavy;
      Font.Height := -15;
      Font.Name := 'Arial';
      Font.Style := [];
      ForceNewPage := False;
      Frame.Width := 0;
{$IfNDef Win32}
      Frame.DrawTop := False;
      Frame.DrawBottom := False;
      Frame.DrawLeft := False;
      Frame.DrawRight := False;
{$EndIf}
      ParentFont := False;
      Ruler := qrrInchesHV;
    end;
    FmtWrite(UnitStream, '    Title:TQRBand;' + CRLF, [nil]);

    if EditTitle.Text <> '' then begin
      QRSysData := TQRSysData.Create(Form);
      with QRSysData do begin
        Parent := Result;
        Name := CheckName(Form,'QRSysData');
        Top := 12;
        Width := GetTextWidth(EditTitle.Text,Font);
        Left := (GetBandWidth - Width) div 2;
        Height := -Font.Height + 4;
        Width := Width + 8; { to prevent }
        AlignToBand := False;
        Data := qrsReportTitle
      end; { with }
      FmtWrite(UnitStream, '    %s:TQRSysData;' + CRLF, [QRSysData.Name]);
    end; { if }

    if CBDateTime.Checked then begin
      QRSysData := TQRSysData.Create(Form);
      with QRSysData do begin
        Parent := Result;
        Name := CheckName(Form,'QRSysData');
        ParentFont := FALSE;
        Font.Height := -13;
        Font.Name := 'Arial';
        Font.Style := [];
        Top := 04;
        S := DateTimeToStr(Now);
        Width := GetTextWidth(S,Font) + 8;
        Left := GetBandWidth - Width;
        Height := -Font.Height + 4;
        AlignToBand := False;
        Alignment := taRightJustify;
        Data := qrsDateTime
      end; { with }
      FmtWrite(UnitStream, '    %s:TQRSysData;' + CRLF, [QRSysData.Name]);
    end; { if }

    if CBPageNumber.Checked then begin
      QRSysData := TQRSysData.Create(Form);
      with QRSysData do begin
        Parent := Result;
        Name := CheckName(Form,'QRSysData');
        ParentFont := FALSE;
        Font.Height := -13;
        Font.Name := 'Arial';
        Font.Style := [];
        Top := 24;
        Text := 'Page #';
        S := '99999';
        Width := GetTextWidth(Text + S,Font) + 8;
        Left := GetBandWidth - Width;
        Height := -Font.Height + 4;
        AlignToBand := False;
        Alignment := taRightJustify;
        Data := qrsPageNumber;
      end; { with }
      FmtWrite(UnitStream, '    %s:TQRSysData;' + CRLF, [QRSysData.Name]);
    end; { if }

    if PictureFileName <> '' then begin
      Image := TImage.Create(Form);
      with Image do begin
        Parent := Result;
        Name := CheckName(Form,'HeaderImage');
        Left := 0;
        Top := 0;
        Width := 105;
        Height := 41;
        Picture.LoadFromFile(PictureFileName);
        Stretch := TRUE;
      end;
      FmtWrite(UnitStream, '    %s:TImage;' + CRLF, [Image.Name]);
    end;
  end;

  function InsertDetailLink(QRXTable:TQRXTable; aIndex:integer):TQRDetailLink;
  begin
    Result := TQRDetailLink.Create(Form);
    with Result do begin
      DataSource := QRXTable.FDataSource;
      Name := CheckName(Form,'QRDetailLink' +  copy(DataSource.Name,11,255));
      Master := GetMasterReport(QRXTable);
      PrintBefore := FALSE;
      LongRec(aDesignInfo).Lo := 4;
      LongRec(aDesignInfo).Hi := QRXTable.FTop;
      Result.DesignInfo := aDesignInfo;
    end;
    FmtWrite(UnitStream, '    %s:TQRDetailLink;' + CRLF,[Result.Name]);
    QRXTable.FController := Result;
  end;

  function GetTableID(TableName:string):string;
  var
    P : byte;
  begin
    Result := TableName;
    P := Pos('.',Result);
    if P <> 0 then
      System.Delete(Result,P,1);
  end;

  function ExtractFieldName(const FullFieldName:string):string;
  var
    P : integer;
  begin
    Result := FullFieldName;
    P := Pos('.',Result);
    if P <> 0 then
      System.Delete(Result,1,P);
  end;

  function GetDataSource(Fields:TStringList; const FullFieldName:string):TDataSource;
  var
    F, P : integer;
    QRXTable : TQRXTable;
    Found : boolean;
    TableID : string;
  begin
    TableID := FullFieldName;
    P := Pos('.',TableID);
    if P <> 0 then
      System.Delete(TableID,P,255);
    F := 1; Found := FALSE;
    while (F <= OutQRXTables.ItemCount) and not Found do begin
      QRXTable := TQRXTable(OutQRXTables.Items[F].Data);
      Found := (GetTableID(QRXTable.GenTable.TableName) = TableID);
      if Found then
        Result := QRXTable.FDataSource
      else
        Inc(F);
    end; { while }
  end;

  function GetFieldWidth(Fields:TStringList; const FullFieldName:string;
    FieldFont:TFont):integer;
  var
    Field : TField;
    S : string;
    F : integer;
  begin
    S := ConstStr('n',10);
    F := Fields.IndexOf(FullFieldName);
    if F <> -1 then begin
      Field := TField(Fields.Objects[F]);
      if (Field is TStringField) then
        S := ConstStr('n',Field.Size + 1)
      else if (Field is TBlobField) then
        S := ConstStr('n',40)
      else if (Field is TDateTimeField) then
        S := DateTimeToStr(Field.AsDateTime)
      else
        S := Field.DisplayText + 'n';
    end;
    Result := GetTextWidth(S,FieldFont);
  end;

  function InsertBand(aBandType:TQRBandType):TQRBand;
  begin
    Result := TQRBand.Create(Form);
    with Result do begin
      Parent := Form;
      Name := CheckName(Form,'QRBand');
      Left := 0;
      Top := CurrBand.Top + 1;
      Width := 1200;
      Height := -Font.Height + 6; { Moacir }
      Align := alTop;
      BandType := aBandType;
      Color := clWhite;
      ForceNewPage := False;
{$IfNDef Win32}
      if BandType = rbColumnHeader then begin
        Frame.DrawBottom := TRUE;
      end else begin
        Frame.DrawBottom := FALSE;
      end;
      Frame.DrawTop := FALSE;
      Frame.DrawLeft := False;
      Frame.DrawRight := False;
{$EndIf}
      Ruler := qrrNone;
    end;
    FmtWrite(UnitStream, '    %s:TQRBand;' + CRLF,
      [Result.Name]);
  end;

  function GetHorWidth(aLabelWidth,aFieldWidth:integer; IsHoriz:boolean):integer;
  begin
    if IsHoriz then
      Result := aLabelWidth + aFieldWidth
    else
      Result := Max(aLabelWidth,aFieldWidth);
  end;

  function InsertLabel(QRBand:TQRBand; const FullFieldName:string;
    aLeft, aTop, aWidth : integer; IsHoriz:boolean):TQRLabel;
  var
    QRLabel : TQRLabel;
    FieldName : string;
  begin
    FieldName := ExtractFieldName(FullFieldName);
    QRLabel := TQRLabel.Create(Form);
    with QRLabel do begin
      Parent := QRBand;
      Name := CheckName(Form,'QRLabel' + FieldName);
      if IsHoriz then
        Caption := FieldName + ': '
      else
        Caption := FieldName;
      AlignToBand := False;
      Font.Assign(LabelFont);
      ParentFont := False;
      Left := aLeft;
      Top := aTop;
      Width := aWidth;
      Height := -Font.Height + 4;
    end;
    FmtWrite(UnitStream, '    %s:TQRLabel;' + CRLF,[QRLabel.Name]);
    Result := QRLabel;
  end;

  procedure InsertField(QRBand:TQRBand; Fields:TStringList; const FullFieldName:string;
    aLeft, aTop, aWidth : integer);
  var
    QRDBText : TQRDBText;
    FieldName : string;
  begin
    FieldName := ExtractFieldName(FullFieldName);
    QRDBText := TQRDBText.Create(Form);
    with QRDBText do begin
      Parent := QRBand;
      Name := CheckName(Form,'QRDBText' + FieldName);
{$IfNDef Win32}
      AlignToBand := False;
{$EndIf}
      DataSource := GetDataSource(Fields,FullFieldName);
      DataField := FieldName;
      Left := aLeft;
      Top := aTop;
      Width := aWidth; { only a "folga" }
      Height := -Font.Height + 4;
    end;
    FmtWrite(UnitStream, '    %s:TQRDBText;' + CRLF,[QRDBText.Name]);
  end;

  procedure InsertLabelOrFields(Fields:TStringList; InsertLabels, InsertFields:boolean;
    Band:TQRBand; aInitialTop, aInitialLeft:integer);
  var
    F : integer;
    aTop, aLeft : integer;
    aLabelWidth, aFieldWidth, aMaxWidth : integer;
  begin { InsertLabelFieldsVer }
    aLeft := aInitialLeft; aTop := aInitialTop;
    for F := 0 to Fields.Count - 1 do begin
      aLabelWidth := GetTextWidth(ExtractFieldName(Fields[F]),LabelFont);
      aFieldWidth := GetFieldWidth(Fields,Fields[F],Form.Font);
      aMaxWidth := Max(aLabelWidth, aFieldWidth);
      if aLeft + aMaxWidth > GetBandWidth then begin
        aLeft := aInitialLeft;
        aTop := aTop - Band.Font.Height + 6;
        Band.Height := Band.Height - Font.Height + 8;
{$IfNDef Win32}
        if Band.BandType in [rbDetail,rbSubDetail] then
          Band.Frame.DrawBottom := TRUE;
{$EndIf}
      end;
      if InsertLabels then InsertLabel(Band,Fields[F],aLeft,aTop,aLabelWidth,FALSE);
      if InsertFields then InsertField(Band,Fields,Fields[F],aLeft,aTop,aFieldWidth);
      aLeft := aLeft + aMaxWidth + Interval;
    end;
  end;

  procedure InsertLabelAndFields(Fields:TStringList;
    Band:TQRBand; aInitialTop, aInitialLeft:integer; IsHoriz:boolean);
  var
    F : integer;
    aTop, aLeft : integer;
    aLabelWidth, aFieldWidth, aMaxWidth : integer;
    LabelInserted : TQRLabel;
  begin { InsertLabelFieldsVer }
    if not IsHoriz then
      Band.Height := Band.Height - Band.Font.Height + 8;
    aLeft := aInitialLeft; aTop := aInitialTop;
    for F := 0 to Fields.Count - 1 do begin
      aLabelWidth := GetTextWidth(ExtractFieldName(Fields[F]) + ': ',LabelFont);
      if IsHoriz then
        aLabelWidth := aLabelWidth + GetTextWidth(': ',LabelFont);
      aFieldWidth := GetFieldWidth(Fields,Fields[F],Form.Font);
      if IsHoriz then
        aMaxWidth := aLabelWidth + aFieldWidth
      else
        aMaxWidth := Max(aLabelWidth,aFieldWidth);
      if aLeft + aMaxWidth > GetBandWidth then begin
        aLeft := aInitialLeft;
        aTop := aTop - Band.Font.Height + 6;
        Band.Height := Band.Height - Font.Height + 8;
        if not IsHoriz then begin
          aTop := aTop - Band.Font.Height + 6;
          Band.Height := Band.Height - Font.Height + 8;
        end;
{$IfNDef Win32}
        if Band.BandType in [rbDetail,rbSubDetail] then
          Band.Frame.DrawBottom := TRUE;
{$EndIf}
      end;
      LabelInserted := InsertLabel(Band,Fields[F],aLeft,aTop,aLabelWidth,IsHoriz);
      if IsHoriz then
        InsertField(Band,Fields,Fields[F],aLeft + aLabelWidth,aTop,aFieldWidth)
      else
        InsertField(Band,Fields,Fields[F],aLeft,aTop - LabelInserted.Font.Height + 8,aFieldWidth);
      aLeft := aLeft + aMaxWidth + Interval;
    end;
  end;

  function GetDetailTable(QRXTable:TQRXTable; const FullFieldName:string):TQRXTable;
  { Return the DetailTable for Field "FieldName". }
  var
    P, F : integer;
    Found : boolean;
    Q : TQRXTable;
    MasterField : string;
    MasterSource : string;
  begin
    F := 1; Found := FALSE; Result := nil;
    while (F <= OutQRXTables.ItemCount) and not Found do begin
      Q := TQRXTable(OutQRXTables.Items[F].Data);
      if Q.GenTable.MasterSource <> nil then begin
        MasterField := Q.GenTable.MasterFields;
        P := Pos(';',MasterField); { to get only first master field }
        if P <> 0 then
          System.Delete(MasterField,P,256);
        MasterSource := GetTableID(QRXTable.GenTable.TableName);
        Found := (MasterSource + '.' + MasterField = FullFieldName)
          and (MasterSource = GetTableID(TTable(Q.GenTable.MasterSource.DataSet).TableName));
        if Found then
          Result := TQRXTable(OutQRXTables.Items[F].Data)
      end;
      Inc(F);
    end;
  end;

  procedure BuildBand(QRXTable:TQRXTable; List:TStringList);
  var
    LocalList : TStringList;
    F : integer;
    DetailTable : TQRXTable;
    FullFieldName : string;
  begin
    LocalList := TStringList.Create;
    try
      QRXTable.GenTable.GetFieldNames(LocalList);
      for F := 0 to LocalList.Count - 1 do begin
        FullFieldName := GetTableID(QRXTable.GenTable.TableName) + '.' + LocalList[F];
        if QRXTable.FSelFields.IndexOf(LocalList[F]) <> -1 then { is selected }
          List.AddObject(FullFieldName,QRXTable.GenTable.FieldByName(LocalList[F]));
        DetailTable := GetDetailTable(QRXTable,FullFieldName);
        if (DetailTable <> nil) and not DetailTable.FIsMDRelation then
          BuildBand(DetailTable,List)
      end;
    finally
      LocalList.Free;
    end; { try }
  end;

  function IsLastBand(QRXTable:TQRXTable):boolean;
  begin
    IsLastBand := (QRXTable = nil) or (GetRealDetailQRXTable(QRXTable) = nil);
  end;

  procedure CheckLinkDetailBand(QRXTable:TQRXTable;Band:TQRBand);
  begin
    if QRXTable.FController is TQRDetailLink then
      TQRDetailLink(QRXTable.FController).DetailBand := Band;
  end;

begin { MainLoop }
  WorkCanvas := TControlCanvas.Create; WorkCanvas.Control := Form;
  try
    LabelFont := TFont.Create;
    LabelFont.Color := clBlack;  { Fonte usada nos rbColumnHeader }
    LabelFont.Height := -13;
    LabelFont.Name := 'Arial';
    LabelFont.Style := [fsBold];
    HasMD := FALSE;
    aNonVisualTop := Form.Height div 2; { Na metade do formulario }
    CurrQRXTable := TQRXTable(OutQRXTables.Items[1].Data);
    InsertTable(CurrQRXTable,aNonVisualTop); Inc(aNonVisualTop,32);
    QuickReport := InsertQuickRep(CurrQRXTable);
    CurrBand := InsertHeader;
    for F := 2 to OutQRXTables.ItemCount do begin
      CurrQRXTable := TQRXTable(OutQRXTables.Items[F].Data);
      InsertTable(CurrQRXTable,aNonVisualTop); Inc(aNonVisualTop,32);
      if CurrQRXTable.FIsMDRelation then begin
        InsertDetailLink(CurrQRXTable,F);
        HasMD := TRUE; { Flag indicando que existe relacionamento Mester/Detalhe }
      end;
    end;

    if not HasMD then begin
      FieldsInBand := TStringList.Create;
      try
        CurrQRXTable := TQRXTable(OutQRXTables.Items[1].Data);
        CurrQRXTable.GenTable.FieldDefs.Update;
        BuildBand(CurrQRXTable,FieldsInBand);
        CurrBand := InsertBand(rbColumnHeader);
        InsertLabelOrFields(FieldsInBand,TRUE,FALSE,CurrBand,IniTop,IniLeft);
        CurrBand := InsertBand(rbDetail);
        InsertLabelOrFields(FieldsInBand,FALSE,TRUE,CurrBand,IniTop,IniLeft);
      finally
        FieldsInBand.Free;
      end; { try }
    end else begin
      FieldsInBand := TStringList.Create;
      FieldsInDetailBand := TStringList.Create;
      try
        CurrQRXTable := TQRXTable(OutQRXTables.Items[1].Data);
        aBandType := rbDetail; { only in first time }
        while CurrQRXTable <> nil do begin
          CurrQRXTable.GenTable.FieldDefs.Update;
          DetailQRXTable := GetRealDetailQRXTable(CurrQRXTable);
          FieldsInBand.Clear;
          FieldsInDetailBand.Clear;
          BuildBand(CurrQRXTable,FieldsInBand);
          if IsLastBand(DetailQRXTable) then begin
            BuildBand(DetailQRXTable,FieldsInDetailBand);
            CurrBand := InsertBand(aBandType);
            CheckLinkDetailBand(CurrQRXTable,CurrBand);
            InsertLabelAndFields(FieldsInBand,CurrBand,IniTop,IniLeft,TRUE);
            CurrBand.Height := CurrBand.Height + 20; { afinar depois }
            InsertLabelOrFields(FieldsInDetailBand,TRUE,FALSE,CurrBand,CurrBand.Height-18,IniLeft+32);
            CurrBand := InsertBand(rbSubDetail);
            CheckLinkDetailBand(DetailQRXTable,CurrBand);
            InsertLabelOrFields(FieldsInDetailBand,FALSE,TRUE,CurrBand,IniTop,IniLeft+32);
            Break;
          end else begin
            CurrBand := InsertBand(aBandType);
            CheckLinkDetailBand(CurrQRXTable,CurrBand);
            InsertLabelAndFields(FieldsInBand,CurrBand,IniTop,IniLeft,FALSE);
          end;
          aBandType := rbSubDetail;
          CurrQRXTable := DetailQRXTable;
        end; { while }
      finally
        FieldsInBand.Free;
        FieldsInDetailBand.Free;
      end; { try }
    end;
    LabelFont.Free;
  finally
    WorkCanvas.Free;
  end;
end;

procedure TFrmQRExpt.Generate(const FileName, FormIdent, UnitIdent: string;
  ToolServices:TIToolServices);
var
  Form : TForm;
  FormStream : TMemoryStream;
  UnitStream : TMemoryStream;
  IFormStream : TIMemoryStream;
  IUnitStream : TIMemoryStream;
begin
{$IfDef Win32}
  Form := TForm.Create(nil);
  Proxies.CreateSubClass(Form,'T' + FormIdent,TForm);
{$Else}
  Form := TProxyForm.CreateAs('T' + FormIdent);
{$EndIf}
  try
    UnitStream := TMemoryStream.Create;
    try
      SourceBuffer := StrAlloc(SourceBufferSize);
      try
        InitModules(Form,UnitStream,FormIdent,UnitIdent);
        MainLoop(Form,UnitStream);
        DoneModules(Form,UnitStream,FormIdent);
        UnitStream.Position := 0;
        FormStream := TMemoryStream.Create;
        try
          FormStream.WriteComponentRes(FormIdent, Form);
          FormStream.Position := 0;
          IFormStream := TIMemoryStream.Create(FormStream);
          try
{$IfDef Win32}
            IFormStream.AddRef;
{$EndIf}
            IUnitStream := TIMemoryStream.Create(UnitStream);
            try
              if ToolServices <> nil then { Apenas para depuracao } begin
{$IfDef Win32}
                IUnitStream.AddRef;
{$EndIf}
                ToolServices.CreateModule(FileName, IUnitStream, IFormStream,
                  [cmAddToProject, cmShowSource, cmShowForm, cmUnNamed,
                  cmMarkModified]);
              end;
            finally
              IUnitStream.OwnStream := FALSE;
              IUnitStream.Free;
            end;
          finally
            IFormStream.OwnStream := FALSE;
            IFormStream.Free;
          end;
        finally
          FormStream.Free;
        end;
      finally
        UnitStream.Free;
      end;
    finally
      StrDispose(SourceBuffer);
    end;
  finally
    Form.Free;
  end;
end;

procedure TFrmQRExpt.FmtWrite(Stream: TStream; Fmt: PChar;
  const Args: array of const);
begin
  if (Stream <> nil) and (SourceBuffer <> nil) then
  begin
    StrLFmt(SourceBuffer, SourceBufferSize, Fmt, Args);
    Stream.Write(SourceBuffer[0], StrLen(SourceBuffer));
  end;
end;

procedure QuickRepExpert(ToolServices: TIToolServices);
var
  QRX : TFrmQRExpt ;
  UnitIdent, FormIdent: {$IfDef Win32} AnsiString; {$Else} string; {$EndIf}
  FileName: TFileName;
begin
  if ToolServices = nil then Exit;
  if ToolServices.GetNewModuleName(UnitIdent, FileName) then begin
    QRX := TFrmQRExpt.Create(Application);
    try
      if QRX.ShowModal = mrOK then begin
        Screen.Cursor := crHourGlass;
        UnitIdent := LowerCase(UnitIdent);
        UnitIdent[1] := Upcase(UnitIdent[1]);
        FormIdent := 'Form' + Copy(UnitIdent, 5, 255);
        QRX.Generate(FileName,FormIdent,UnitIdent,ToolServices);
      end;
    finally
      Screen.Cursor := crDefault;
      QRX.Free;
    end; { try }
  end; { if }
end;

end.
